"""
Collect library
functions that are used to collect information from the network to source nodes.
"""

from phyelds.calculus import neighbors, aggregate, remember
from phyelds.libraries.device import local_id


@aggregate
def find_parent(potential: float) -> any:
    """
    Find the parent of a path giving a potential field
    :param potential:
    :return:
    """
    neighbors_potential = neighbors(potential)
    min_value = min(neighbors_potential.data.items(), key=lambda x: x[1])
    if min_value[1] >= potential:
        return None
    return min_value[0]


@aggregate
def collect_with(potential, local, accumulation):
    """
    TODO
    :param potential:
    :param local:
    :param accumulation:
    :return:
    """
    collections = remember(local)
    n_collections = neighbors(collections)
    parents = neighbors(find_parent(potential))
    zip_operation = zip(parents, n_collections)
    operations = local
    for parent, value in zip_operation:
        if local_id() == parent:
            operations = accumulation(operations, value)
    return collections.update(operations)


@aggregate
def count_nodes(potential):
    """
    TODO
    :param context:
    :param potential:
    :return:
    """
    return collect_with(potential, 1, lambda x, y: x + y)


@aggregate
def sum_values(potential, local):
    """
    TODO
    :param context:
    :param potential:
    :param local:
    :return:
    """
    return collect_with(potential, local, lambda x, y: x + y)


@aggregate
def collect_or(potential, local):
    """
    TODO
    :param context:
    :param potential:
    :param local:
    :return:
    """
    return collect_with(potential, local, lambda x, y: x or y)



"""
phyelds: A Python library for aggregate computing system with a pythonic interface.

This library provides a set of tools for building and running aggregate computing systems.

A global engine is used to manage the state and message passing between different contexts.
In order to use it, you should:
1. Call `engine.setup(messages, id, state)` to initialize the engine with the current context.
2. Call an aggregate script
3. Call `engine.cooldown()` to reset the engine state and get the messages to send.
4. Store the state for the next iteration.

"""
from contextvars import ContextVar
from phyelds.abstractions import Engine

engine: ContextVar[Engine] = ContextVar("engine")



"""
Engine interface: how to implement an aggregate computing engine.
"""
from abc import ABC
from dataclasses import dataclass
from typing import List, Dict, Any


class NodeContext(ABC):
    """
    Abstract base class for the node context. This class should be implemented by the user.
    """

    def __init__(self, node_id: int, sensors: dict[str, any] = None):
        """
        Initialize the node context with the node id and sensors.
        :param node_id: The id of the node.
        :param sensors: The sensors of the node.
        """
        self.node_id = node_id
        self.sensors = sensors  # map sensor name to sensor value

    def sense(self, sensor_name: str) -> any:
        """
        Get the value of the sensor.
        :param sensor_name: The name of the sensor.
        :return: The value of the sensor.
        """
        return self.sensors.get(sensor_name)

    def position(self) -> any:
        """
        Get the position of the node.
        :return: The position of the node.
        """
        return self.sensors.get("position")


@dataclass
class EngineState:
    """
    Inner state of the engine used to process aggregate computation.
    """
    stack: List[str] = None
    state_trace: Dict[str, Any] = None
    count_stack: List[int] = None
    to_send: Dict[str, Any] = None
    messages: Dict[int, Dict[str, Any]] = None
    node_id: int = 0
    reads: set = None

    def __post_init__(self):
        self.stack = [] if self.stack is None else self.stack
        self.state_trace = {} if self.state_trace is None else self.state_trace
        self.count_stack = [0] if self.count_stack is None else self.count_stack
        self.to_send = {} if self.to_send is None else self.to_send
        self.messages = {} if self.messages is None else self.messages
        self.reads = set() if self.reads is None else self.reads


class Engine(ABC):
    """
    Abstract base class for the engine. This class should be implemented by the user.
    """

    def __init__(self):
        self.node_context: NodeContext | None = None
        self.engine_state: EngineState | None = None

    def setup(
        self, node_context: NodeContext, messages: dict[int, dict[str, any]], state=None
    ) -> "Engine":
        """
        Setup the engine with the current context.
        :param node_context: the current context of the engine.
        :param messages: The messages to send.
        :param state: The state of the engine.
        """

    def enter(self, name: str) -> None:
        """
        Enter a new context.
        :param name: The name of the context.
        """

    def exit(self) -> None:
        """
        Exit the current context.
        """

    def current_path(self) -> list[str]:
        """
        Get the current path of the engine.
        :return: The current path.
        """

    def write_state(self, value: any, stack: list[str]) -> None:
        """
        Write the state of the engine.
        :param value: The value to write.
        :param stack: The stack of the engine.
        """

    def read_state(self, stack: list[str]) -> any:
        """
        Read the state of the engine.
        :param stack: The stack of the engine.
        :return: The value of the state.
        """

    def state_trace(self) -> dict[str, any]:
        """
        Get the state trace of the engine.
        :return: The state trace.
        """

    def forget(self, _self_path):
        """
        Forget the value store in the engine.
        :param _self_path:
        :return:
        """

    def send(self, data: any) -> None:
        """
        Send data to the engine.
        :param data: The data to send.
        """

    def aligned(self) -> list[int]:
        """
        Get the aligned ids.
        :return: The aligned ids.
        """

    def aligned_values(self, path: list[str]) -> dict[int, any]:
        """
        Get the aligned values.
        :param path: The path of the values.
        :return: The aligned values.
        """

    def cooldown(self) -> dict[str, Any]:
        """
        Cooldown the engine.
        :return:
        """



"""
Device library:
Set of function used to get the device information.
"""
from phyelds import engine


def local_id():
    """
    Get the local id of the device.
    :return:
    """
    return engine.get().node_context.node_id


def context():
    """
    Get the context of the device.
    :return:
    """
    return engine.get().node_context


def sense(sensor: str) -> any:
    """
    Get the value of the sensor.
    :param sensor: The name of the sensor.
    :return: The value of the sensor.
    """
    return engine.get().node_context.sensors[sensor]


def local_position():
    """
    Get the position of the device.
    :return: The position of the device.
    """
    return engine.get().node_context.position()



"""
This module contains functions to calculate distances between nodes in the neighborhood.
These functions may be used by other library to compute system-wise properties
"""

from phyelds.calculus import neighbors, aggregate
from phyelds.data import Field
from phyelds.libraries.device import local_id


@aggregate
def neighbors_distances(position):
    """
    Get the distances to the neighbors from the current node.
    :param position: the current node position
    :return: the field representing the distances to the neighbors
    """
    positions = neighbors(position)
    x, y = position
    distances = {}
    for node_id, pos in positions.data.items():
        # pos are x, y tuples
        n_x, n_y = pos
        distances[node_id] = ((x - n_x) ** 2 + (y - n_y) ** 2) ** 0.5
    return Field(distances, local_id())


@aggregate
def hops_distance():
    """
    Get the hops distance to the neighbors from the current node.
    :return: the field representing the hops distance to the neighbors
    """
    distances = neighbors(1)
    distances.data[local_id()] = 0
    return distances



"""
Render the nodes in the simulator's environment.
"""

from typing import Tuple

from matplotlib import pyplot as plt

from phyelds.simulator import Simulator


class Link:
    """
    A class to represent a link between two nodes.
    The link is represented as a tuple of two nodes.
    Note!
    Link(node1, node2) == Link(node2, node1)
    """

    def __init__(self, node1: Tuple[float, ...], node2: Tuple[float, ...]):
        self.node1 = node1
        self.node2 = node2

    def __hash__(self):
        return hash(frozenset((self.node1, self.node2)))

    def __eq__(self, other):
        if isinstance(other, Link):
            return frozenset((self.node1, self.node2)) == frozenset(
                (other.node1, other.node2)
            )
        return False


def render_sync(simulator: Simulator, color_from: str = None):
    """
    Render the nodes in the simulator's environment.
    """
    positions = [node.position for node in simulator.environment.nodes.values()]
    x, y = zip(*positions)
    # plot also the neighbors, avoiding duplicates
    all_neighbors_tuple = set()
    for node in simulator.environment.nodes.values():
        neighbors = node.get_neighbors()
        for neighbor in neighbors:
            all_neighbors_tuple.add(Link(node.position, neighbor.position))
    for link in all_neighbors_tuple:
        plt.plot(
            [link.node1[0], link.node2[0]],
            [link.node1[1], link.node2[1]],
            "r--",
            alpha=0.1,
        )
    # take the colors from the node data
    if color_from:
        colors = [
            node.data.get(color_from, "blue")
            for node in simulator.environment.nodes.values()
        ]
        plt.scatter(x, y, c=colors)
    else:
        plt.scatter(x, y, c="blue")
    # Add node IDs as text labels
    for node in simulator.environment.nodes.values():
        plt.text(
            node.position[0] + 0.02,
            node.position[1],
            str(node.id),
            fontsize=8,
            ha="center",
            va="center",
            bbox={"facecolor": "white", "alpha": 0.1, "edgecolor": "none"},
        )
    plt.title("Node Positions")
    plt.xlabel("X Position")
    plt.ylabel("Y Position")
    plt.axis("equal")
    plt.show()
    simulator.schedule_event(
        1.0, render_sync, simulator, color_from
    )  # Schedule next render



"""
Neighborhood functions for the phyelds simulator.
A neighborhood function is simply a function that takes a node and a list of all nodes,
and returns a list of neighbors.
"""

from typing import List

from phyelds.simulator import Node


def radius_neighborhood(radius: float):
    """Create a neighborhood function that includes nodes within a certain radius"""

    def neighborhood_func(node: Node, all_nodes: List[Node]) -> List[Node]:
        neighbors = []
        for other in all_nodes:
            if other.id == node.id:  # Skip self
                continue

            # Calculate Euclidean distance
            distance = (
                sum((a - b) ** 2 for a, b in zip(node.position, other.position)) ** 0.5
            )

            if distance <= radius:
                neighbors.append(other)
        return neighbors

    return neighborhood_func


def k_nearest_neighbors(k: int):
    """Create a neighborhood function that includes the k nearest nodes"""

    def neighborhood_func(node: Node, all_nodes: List[Node]) -> List[Node]:
        others = [n for n in all_nodes if n.id != node.id]
        if not others:
            return []

        # Calculate distances
        distances = [
            (
                sum((a - b) ** 2 for a, b in zip(node.position, other.position)) ** 0.5,
                other,
            )
            for other in others
        ]

        # Sort by distance and take k nearest
        distances.sort()
        return [other for _, other in distances[:k]]

    return neighborhood_func


def full_neighborhood(node: Node, all_nodes: List[Node]) -> List[Node]:
    """Include all nodes as neighbors"""
    return [n for n in all_nodes if n.id != node.id]



"""
This module contains utility functions and classes for various purposes.
"""


def min_with_default(iterable, default=None):
    """
    Returns the minimum value from an iterable, or a default value if the iterable is empty.
    :param iterable: a list of values (or any iterable)
    :param default: the default value to return if the iterable is empty
    :return: the minimum value from the iterable or the default value
    """
    if not iterable:
        return default
    if isinstance(iterable, dict):
        return min(iterable.values())
    if isinstance(iterable, (list, tuple, set, range)):
        return min(iterable)
    return default



"""
Diffusion (information) library for phyelds
This library provides functions for diffusion of information in a network.
It includes functions for calculating distances and to braodcast information
"""
from phyelds.calculus import aggregate, remember, neighbors
from phyelds.data import Field
from phyelds.libraries.utils import min_with_default


@aggregate
def distance_to(source: bool, distances: Field) -> float:
    """
    Calculate the distance to a source node in the network.
    :param source:
    :param distances:
    :return:
    """
    gradient = remember(float("inf"))
    neighbors_gradients = neighbors(gradient) + distances
    return gradient.update(
        0.0
        if source
        else min_with_default(neighbors_gradients.exclude_self(), float("inf"))
    )


def cast_from(source: bool, data: any, accumulation: callable, distances: Field) -> any:
    """
    Cast information from a source node to its neighbors.
    :param source: the source node to cast from
    :param data: the data to cast
    :param accumulation: the function to accumulate the data
    :param distances: the distances against the neighbors
    :return: the data cast
    """
    cast_area = remember(data)
    potential = distance_to(source, distances)
    neighbors_value = neighbors(cast_area)
    # neighbors potential
    neighbors_potential = neighbors(potential)
    # take the value from the minimum potential
    values = zip(neighbors_potential, neighbors_value)
    # select the minimum potential
    _, result = min(values, key=lambda x: x[0])
    if source:
        return cast_area.update(data)
    return cast_area.update(accumulation(result))


@aggregate
def broadcast(source: bool, data: any, distances: Field) -> any:
    """
    Broadcast information to the network from a source node.
    :param source: the source node to broadcast from
    :param data: the data to spread
    :param distances: the distances again the neighbors
    :return: the data spread
    """
    return cast_from(source, data, lambda x: x, distances)



"""
Internal state class used to manage the state of the system (namely `rep` of field calculus).
"""

from typing import Any, Dict, Optional, List
import wrapt
from phyelds.abstractions import Engine


class Field:
    """
    Field class used to manage the interactions of between nodes (namely `nbr` of field calculus).
    It provides methods to perform operations on the field, such as addition,
    subtraction, multiplication, and division.
    You should never use it directly, but rather use the `neighbors` function
    """

    def __init__(self, data: Dict[int, Any], node_id: int) -> None:
        self.data = dict(sorted(data.items()))
        self._iter_index: Optional[int] = None
        self._iter_keys: Optional[List[int]] = None
        self.node_id = node_id

    def __iter__(self) -> "Field":
        self._iter_index = 0
        self._iter_keys = sorted(self.data.keys())
        return self

    def __next__(self) -> Any:
        if self._iter_keys is None or self._iter_index is None:
            raise StopIteration

        if self._iter_index >= len(self._iter_keys):
            self._iter_index = None
            self._iter_keys = None
            raise StopIteration

        key = self._iter_keys[self._iter_index]
        value = self.data[key]
        self._iter_index += 1
        return value

    def exclude_self(self) -> dict[int, Any]:
        """
        Exclude the current node from the field.
        :return:  A new Field object with the current node excluded.
        """
        to_return = self.data.copy()
        to_return.pop(self.node_id, None)
        return to_return

    def local(self):
        """
        Get the local value of the current node.
        :return:
        """
        return self.data.get(self.node_id, None)

    def select(self, field) -> list:
        """
        Select the values from the field that are present in the current field.
        :param field: The field to select from.
        :return:  A list of values from the current field that are present in the given field.
        """
        return [
            self.data[k] for k in self.data.keys() & field.data.keys() if field.data[k]
        ]

    # Helper method to apply binary operations
    def _apply_binary_op(self, other, op):
        if isinstance(other, Field):
            return Field(
                {
                    k: op(self.data[k], other.data[k])
                    for k in self.data.keys() & other.data.keys()
                },
                self.node_id,
            )
        return Field({k: op(v, other) for k, v in self.data.items()}, self.node_id)

    def __add__(self, other):
        return self._apply_binary_op(other, lambda a, b: a + b)

    def __sub__(self, other):
        return self._apply_binary_op(other, lambda a, b: a - b)

    def __mul__(self, other):
        return self._apply_binary_op(other, lambda a, b: a * b)

    def __truediv__(self, other):
        return self._apply_binary_op(other, lambda a, b: a / b)

    def __mod__(self, other):
        return self._apply_binary_op(other, lambda a, b: a % b)

    def __pow__(self, other):
        return self._apply_binary_op(other, lambda a, b: a**b)

    def __floordiv__(self, other):
        return self._apply_binary_op(other, lambda a, b: a // b)

    def __and__(self, other):
        return self._apply_binary_op(other, lambda a, b: a & b)

    def __or__(self, other):
        return self._apply_binary_op(other, lambda a, b: a | b)

    def __xor__(self, other):
        return self._apply_binary_op(other, lambda a, b: a ^ b)

    def __invert__(self):
        return Field({k: ~v for k, v in self.data.items()}, self.node_id)

    def __lt__(self, other):
        return self._apply_binary_op(other, lambda a, b: a < b)

    def __le__(self, other):
        return self._apply_binary_op(other, lambda a, b: a <= b)

    def __gt__(self, other):
        return self._apply_binary_op(other, lambda a, b: a > b)

    def __str__(self):
        """String representation of the field."""
        return self.__repr__()

    def __repr__(self):
        """String representation of the field."""
        return f"Field (id: {self.node_id}) -- data: {self.data} -- local: {self.local()}"


class State(wrapt.ObjectProxy):
    """
    A wrapper class that delegates operations to the underlying value
    while maintaining state management functionality.
    """

    def __init__(self, default: Any, path: List, engine: Engine):
        self.__wrapped__ = default  # then it will be updated
        state = engine.read_state(path)
        if state is None:
            value = default
            engine.write_state(default, path)
        else:
            value = state
        super().__init__(value)
        self._self_path = list(path)
        self._self_engine = engine

    @property
    def value(self) -> Any:
        """Get the current value."""
        return self.__wrapped__

    def update(self, new_value: Any) -> Any:
        """Update the stored value."""
        if isinstance(new_value, State):
            new_value = new_value.value
        self._self_engine.write_state(new_value, self._self_path)
        self.__wrapped__ = new_value
        return self

    def update_fn(self, fn: callable) -> Any:
        """Update the stored value using a function."""
        self.__wrapped__ = fn(self.__wrapped__)
        self._self_engine.write_state(self.__wrapped__, self._self_path)
        return self

    def forget(self):
        """Forget the stored value."""
        self._self_engine.forget(self._self_path)
        self.__wrapped__ = None

    def __str__(self):
        """String representation of the state."""
        return str(self.__wrapped__)

    def __repr__(self):
        """String representation of the state."""
        return f"State: {repr(self.__wrapped__)}"

    def __copy__(self):
        """Create a shallow copy of the state."""
        return State(self.__wrapped__, self._self_path, self._self_engine)

    def __deepcopy__(self, memo):
        """Create a deep copy of the state."""
        return self.__copy__()

    def __reduce__(self):
        """Reduce the state for pickling."""
        return self.__wrapped__

    def __reduce_ex__(self, protocol):
        """Reduce the state for pickling."""
        return self.__reduce__()



"""
Leader election functionality (self-stabilizing) using UUIDs.
"""
import random
from phyelds.data import Field
from phyelds.libraries.device import local_id
from phyelds.libraries.spreading import distance_to
from phyelds.libraries.utils import min_with_default
from phyelds.calculus import aggregate, remember, neighbors


@aggregate
def elect_leader(area: float, distances: Field) -> bool:
    """
    Elect a leader in the network using a random UUID.
    :param area: the area of the network
    :param distances: the distances to the neighbors
    :return: the current id of the leader
    """
    result = breaking_using_uids(random_uuid(), area, distances)
    # Return None if no leader was elected (infinite distance), otherwise return the leader ID
    return result[1] == local_id() and result[0] != (float("inf"))


@aggregate
def random_uuid():
    """
    Generate a random UUID for the node.
    :return: the random UUID
    """
    value = remember(random.random())
    return value, local_id()


@aggregate
def breaking_using_uids(uid, area: float, distances: Field):
    """
    Break the symmetry using the UUID of the node.
    :param uid: the UUID of the node
    :param area: the area of the network
    :param distances: the distances to the neighbors
    :return: the current id of the leader
    """
    # get the minimum value of the neighbors
    lead = remember(uid)
    # get the minimum value of the neighbors
    potential = distance_to(lead == uid, distances)
    new_lead = distance_competition(potential, area, uid, lead, distances)
    # if the new lead is the same, return the uid
    return lead.update(new_lead)


@aggregate
def distance_competition(
    current_distance, area: float, uid, lead, distances: Field
):
    """
    Compare the distance of the current node with the distance of the neighbors.
    It leverages the fact that the tuple is a symmetric breaker.
    :param current_distance: current distance to the lead
    :param area: the maximum distance to the lead
    :param uid: the local id of the node
    :param lead: the current lead
    :param distances: the distances to the neighbors
    :return: the new lead
    """
    inf = (float("inf"), uid[1])
    # neighbors lead
    neighbors_lead = neighbors(lead)
    condition = (neighbors(current_distance) + distances) < (0.5 * area)
    # filter the one that have the condition
    lead = neighbors_lead.select(condition)
    # take the minimum value, but the comparator just consider both values of the tuple
    lead = min_with_default(lead, uid)
    if current_distance > area:
        return uid
    if current_distance >= (0.5 * area):
        return inf
    return lead



"""
This module contains the event to run the aggregate program in the simulator
"""
from abc import ABC

from phyelds.internal import MutableEngine

from phyelds import engine
from phyelds.abstractions import NodeContext
from phyelds.data import State
from phyelds.simulator import Simulator, Node


class SimulatorNodeContext(NodeContext, ABC):
    """
    A class to represent the context of a node in the simulator.
    This class is used to pass the node's data and messages to the aggregate program.
    """

    @staticmethod
    def from_node(node: Node):
        """
        Create a SimulatorNodeContext from a Node.
        :param node: The node to create the context from.
        :return: The SimulatorNodeContext.
        """
        sensors = {
            "position": node.position,
            **node.data,
        }
        return SimulatorNodeContext(node.id, sensors)


def aggregate_program_runner(
    simulator: Simulator, time_delta: float, node: Node, program: callable
):
    """
    Run the program for a node.
    """
    # get neighbors
    all_neighbors = simulator.environment.get_neighbors(node)
    # take the messages from the neighbors, create a dict like id -> messages (that is a dict)
    neighbors_messages = {
        neighbor.id: neighbor.data.get("messages", {}) for neighbor in all_neighbors
    }
    engine.set(MutableEngine())
    engine.get().setup(
        SimulatorNodeContext.from_node(node),
        neighbors_messages,
        node.data.get("state", {})
    )
    result = program()
    if isinstance(result, State):
        result = result.value
    node.data["result"] = result
    node.data["messages"] = engine.get().cooldown()
    node.data["state"] = engine.get().state_trace()
    simulator.schedule_event(
        time_delta, aggregate_program_runner, simulator, time_delta, node, program
    )


def schedule_program_for_all(simulator: Simulator, frequency: float, program: callable):
    """
    Schedule the program for all nodes in the simulator.
    :param simulator: The simulator to schedule the program for.
    :param frequency: The frequency to run the program.
    :param program: The program to run.
    """
    for node in simulator.environment.nodes.values():
        simulator.schedule_event(
            frequency, aggregate_program_runner, simulator, frequency, node, program
        )



"""
Sent of actions related to node movements and interactions in the simulator
"""
import random
from typing import Tuple

from phyelds.simulator import Simulator, Node


def move_with_velocity(
        simulator: Simulator, delta_time: float, node: Node, velocity: float
):
    """
    Move the node with a given velocity.
    """
    # Update the node's position based on its velocity and the delta time
    (x, y) = node.position
    (vx, vy) = velocity
    new_position = (x + vx * delta_time, y + vy * delta_time)
    node.update(new_position)
    simulator.schedule_event(
        delta_time, move_with_velocity, simulator, delta_time, node, velocity
    )


def gaussian_movement(
    simulator: Simulator, node: Node, mean: Tuple[float, ...], stddev: float
):
    """
    Move a node according to a Gaussian distribution.
    """
    new_position = tuple(random.gauss(mean[i], stddev) for i in range(len(mean)))
    node.update(new_position)
    # next schedule the event
    simulator.schedule_event(1.0, gaussian_movement, simulator, node, mean, stddev)



"""
A simple simulator for aggregate computing systems in phyelds.
This simulator allows you to create nodes, set their positions and data,
and define their neighborhoods.
It also provides a way to schedule events and run the simulation.
"""

from typing import Dict, Callable, Any, Optional, Tuple, List
import heapq
import uuid


class Node:
    """
    A class to represent a node in the simulation.
    """

    def __init__(
        self, position: Tuple[float, ...], data: Any = None, node_id: any = None
    ):
        if node_id == 0:
            self.id = 0
        else:
            self.id = node_id or str(uuid.uuid4())
        self.position = position
        if data is None:
            data = {}
        self.data = data
        self.environment = None

    def update(
        self, new_position: Optional[Tuple[float, ...]] = None, new_data: Any = None
    ):
        """Update node position and/or data"""
        if new_position is not None:
            self.position = new_position
        if new_data is not None:
            self.data = new_data
        if self.environment:
            self.environment.node_updated(self)

    def get_neighbors(self):
        """Get neighboring nodes from the environment"""
        if self.environment:
            return self.environment.get_neighbors(self)
        return []


class Environment:
    """
    A class to represent the environment in which nodes exist.
    It manages the nodes and their neighborhoods.
    It also provides a way to set the neighborhood function.
    """

    def __init__(
        self, neighborhood_function: Callable[[Node, List[Node]], List[Node]] = None
    ):
        self.nodes: Dict[any, Node] = {}
        self.neighborhood_function = neighborhood_function or self.no_neighbors

    def add_data_for_all_nodes(self, data: Dict[str, Any]):
        """Add data to all nodes in the environment"""
        for node in self.nodes.values():
            node.data.update(data)

    def node_list(self) -> List[Node]:
        """Return a list of all nodes in the environment"""
        return list(self.nodes.values())

    def add_node(self, node: Node):
        """Add a node to the environment"""
        self.nodes[node.id] = node
        node.environment = self

    def remove_node(self, node_id: str):
        """Remove a node from the environment"""
        if node_id in self.nodes:
            self.nodes[node_id].environment = None
            del self.nodes[node_id]

    def node_updated(self, node: Node):
        """Called when a node is updated"""

    def set_neighborhood_function(self, func: Callable[[Node, List[Node]], List[Node]]):
        """Set the function that determines neighborhoods"""
        self.neighborhood_function = func

    def get_neighbors(self, node: Node) -> List[Node]:
        """Get neighbors for a node using the neighborhood function"""
        return self.neighborhood_function(node, list(self.nodes.values()))

    @staticmethod
    def no_neighbors(
            node: Node,  # pylint: disable=unused-argument
            all_node: List[Node]  # pylint: disable=unused-argument
    ) -> List[Node]:
        """Default neighborhood function (no neighbors)"""
        return []


class Event:
    """
    A class to represent an event in the simulation.
    An event has a time, an action (function), and optional arguments.
    """

    def __init__(self, time: float, action: Callable[..., None], *args, **kwargs):
        self.time = time
        self.action = action
        self.args = args
        self.kwargs = kwargs

    def execute(self):
        """Execute the event's action"""
        return self.action(*self.args, **self.kwargs)

    def __lt__(self, other):
        """For priority queue ordering"""
        return self.time < other.time


class Simulator:
    """
    A class to represent the simulator for a simple aggregate computing system.
    """
    def __init__(self):
        self.event_queue = []
        self.current_time = 0.0
        self.running = False
        self.environment = Environment()

    def schedule_event(
        self, time_delta: float, action: Callable[..., None], *args, **kwargs
    ):
        """Schedule an event to occur after time_delta"""
        event_time = self.current_time + time_delta
        event = Event(event_time, action, *args, **kwargs)
        heapq.heappush(self.event_queue, event)
        return event

    def run(self, until_time: Optional[float] = None):
        """Run the simulation until the specified time or until no more events"""
        self.running = True

        while self.running and self.event_queue:
            event = heapq.heappop(self.event_queue)

            if until_time is not None and event.time > until_time:
                heapq.heappush(self.event_queue, event)  # Put the event back
                break

            self.current_time = event.time
            event.execute()

        self.running = False

    def stop(self):
        """Stop the simulation"""
        self.running = False

    def cancel_event(self, event: Event):
        """Cancel a scheduled event"""
        if event in self.event_queue:
            self.event_queue.remove(event)
            heapq.heapify(self.event_queue)
        else:
            raise ValueError("Event not found in the queue")

    def reset(self):
        """Reset the simulator"""
        self.event_queue = []
        self.current_time = 0.0
        self.running = False
        self.environment = Environment()

    def create_node(
        self, position: Tuple[float, ...], data: Any = None, node_id=None
    ) -> Node:
        """Helper method to create and add a node to the environment"""
        node = Node(position, data, node_id)
        self.environment.add_node(node)
        return node



"""
Core syntax for phyelds.
It exposes a set of functions that are used to create and manipulate fields,
as well as to manage the state of the system.
@aggregate is a decorator that marks a function as an aggregate function,
Example:
@aggregate
def my_function():
    # do something
    return result

Then, there is the core syntax of phyelds:
- remember: used to remember a value across iterations.
- neighbors: used to get the neighbors of a node.
- neighbors_distances: used to get the distances to the neighbors.
"""

from phyelds import engine
from phyelds.calculus.align import AlignContext
from phyelds.data import State, Field


def aggregate(func):
    """
    A decorator for aggregate functions,
    namely each function that is called in the context of a field.
    You can use it in the following way:
    @aggregate
    def my_function():
        # do something
        return result
    """

    def wrapper(*args, **kwargs):
        engine.get().enter(func.__name__)
        result = func(*args, **kwargs)
        engine.get().exit()
        return result

    return wrapper


@aggregate
def remember(init):
    """
    One of the main operator of phyelds TODO
    :param init:
    :return:
    """
    return State(init, engine.get().current_path(), engine.get())


@aggregate
def neighbors(value):
    """
    Get the `value` of the neighbors from the current node.
    Example:
    neighbors(context.data["temperature"]) // returns the temperature of the neighbors
    :param value: used to query the neighbors.
    :return: the field representing this value
    """
    if isinstance(value, Field):
        raise TypeError("Field is not supported as a value")
    engine.get().send(value)
    values = engine.get().aligned_values(engine.get().current_path())
    values[engine.get().node_context.node_id] = value
    return Field(values, engine.get().node_context.node_id)


def align(name: str):
    """
    Used to align a part of the code with the current context,
    creating different non communicating zones
    :param name: what you would like to align on
    :return: the context
    """
    return AlignContext(name)


def align_right():
    """
    Typically used in if statements to align the code
    Example:
    if condition:
        with align_left():
            # do something
    else:
        with align_right():
            # do something else
    :return:
    """
    return align("left")


def align_left():
    """
    Typically used in if statements to align the code
    See align_right
    """
    return align("right")



"""
Engine is the core of the phyelds library. It manages the state and message passing
between different contexts. It provides methods to enter and exit contexts, send messages,
and manage the state of the system.
"""

from typing import Dict, List, Any, Optional

from phyelds.abstractions import Engine, NodeContext, EngineState


class MutableEngine(Engine):
    """
    MutableEngine is the responsible for managing the state and message passing
    in an aggregate computing system. It provides methods to enter and exit contexts,
    send messages, and manage the state of the system.
    """
    def __init__(self):
        super().__init__()
        self.engine_state = EngineState()

    def setup(
        self, node_context: NodeContext, messages=None, state=None
    ) -> Engine:
        if messages is None:
            messages = {}
        if state is None:
            state = {}
        self.node_context = node_context
        self.engine_state = EngineState(
            stack=[],
            state_trace=state.copy(),  # Copy the state to avoid modifying the original
            count_stack=[0],  # Reset counter stack
            to_send={},
            messages=messages,
            node_id=node_context.node_id,
            reads=set(),
        )
        return self

    def enter(self, name: str) -> None:
        counter: int = self.engine_state.count_stack[-1]
        self.engine_state.count_stack[-1] += 1
        self.engine_state.stack.append(f"{name}@{counter}")
        self.engine_state.count_stack.append(0)

    def forget(self, stack: List[str]) -> None:
        if str(stack) in self.engine_state.state_trace:
            del self.engine_state.state_trace[str(stack)]

    def exit(self) -> None:
        if self.engine_state.stack:
            self.engine_state.stack.pop()
            self.engine_state.count_stack.pop()

    def current_path(self) -> list[str]:
        return self.engine_state.stack.copy()

    def write_state(self, value: Any, stack: List[str]) -> None:
        self.engine_state.state_trace[str(stack)] = value

    def read_state(self, stack: List[str]) -> Optional[Any]:
        self.engine_state.reads.add(str(stack))
        return self.engine_state.state_trace.get(str(stack))

    def send(self, data: Any) -> None:
        self.engine_state.to_send[str(self.engine_state.stack)] = data

    def aligned(self) -> List[int]:
        aligned: List[int] = []
        for node_id in self.engine_state.messages:
            if str(self.engine_state.stack) in self.engine_state.messages[node_id]:
                aligned.append(node_id)

        return aligned

    def aligned_values(self, path: List[str]) -> Dict[int, Any]:
        aligned: List[int] = self.aligned()
        path_str: str = str(path)
        aligned_values: Dict[int, Any] = {}
        # take the values of the given path
        for node_id in aligned:
            if path_str in self.engine_state.messages[node_id]:
                aligned_values[node_id] = self.engine_state.messages[node_id][path_str]
        return aligned_values

    def state_trace(self) -> Dict[str, Any]:
        """
        Get the state trace of the engine.
        :return: The state trace.
        """
        return self.engine_state.state_trace.copy()

    def cooldown(self) -> Dict[str, Any]:
        flatten_messages: Dict[str, Any] = {}
        for key in self.engine_state.to_send:
            value = self.engine_state.to_send[key]
            flatten_messages[key] = value
        # get state that were not read
        for key in self.engine_state.state_trace.copy():
            if key not in self.engine_state.reads:
                del self.engine_state.state_trace[key]
        self.engine_state.to_send = {}
        self.engine_state.stack = []
        self.engine_state.count_stack = []  # Reset counter stack
        self.engine_state.messages = {}
        return flatten_messages



"""
This module contains various functions to generate different types of node distributions
in a simulation environment.
These functions can be used to create specific patterns or random distributions of nodes
for testing and simulation purposes.
"""

import math
import random

from phyelds.simulator import Simulator


def grid_generation(simulator: Simulator, width: int, height: int, spacing: float):
    """
    Generate a grid of nodes in the simulator's environment.
    """
    for x in range(0, width):
        for y in range(0, height):
            position = (x * spacing, y * spacing)
            simulator.create_node(position, node_id=x * height + y)


def deformed_lattice(
    simulator: Simulator,
    width: int,
    height: int,
    spacing: float,
    deformation_factor: float,
):
    """
    Generate a deformed lattice of nodes in the simulator's environment.
    """
    for x in range(0, width):
        for y in range(0, height):
            # Randomly deform the position
            dx = random.uniform(-deformation_factor, deformation_factor)
            dy = random.uniform(-deformation_factor, deformation_factor)
            position = (x * spacing + dx, y * spacing + dy)
            simulator.create_node(position, node_id=x * height + y)


def random_walk(simulator: Simulator, num_steps: int, step_size: float):
    """
    Perform a random walk and create nodes at each step.
    """
    position = (0.0, 0.0)
    for _ in range(num_steps):
        dx = random.uniform(-step_size, step_size)
        dy = random.uniform(-step_size, step_size)
        position = (position[0] + dx, position[1] + dy)
        simulator.create_node(position)


def random_in_circle(simulator: Simulator, num_nodes: int, radius: float):
    """
    Generate nodes randomly distributed within a circle.
    """
    for index in range(num_nodes):
        angle = random.uniform(0, 2 * math.pi)
        r = random.uniform(0, radius)
        x = r * math.cos(angle)
        y = r * math.sin(angle)
        simulator.create_node((x, y), node_id=index)



"""
A group of functions based on the notion of time.
"""

from phyelds.calculus import aggregate, remember


@aggregate
def counter():
    """
    Simple counter function that counts the number of times it is called.
    :return: a counter that counts the number of times it is called.
    """
    return remember(0).update_fn(lambda x: x + 1)



"""
Alignment context manager for phyelds.

"""

from phyelds import engine


class AlignContext:
    """
    Context manager for alignment in phyelds.
    This context manager is used to align the state of the engine with the
    current context. It should be used in the following way:
    with AlignContext("context_name"):
        # do something
        pass

    """
    def __init__(self, name: str):
        self.name = name

    def __enter__(self):
        engine.get().enter(self.name)
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        engine.get().exit()
